package be.iminds.ilabt.jfed.restapi.resource;

import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.CredentialException;
import be.iminds.ilabt.jfed.lowlevel.SimpleGeniUserWithSpeaksFor;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.restapi.JFedRestApiConfiguration;
import be.iminds.ilabt.jfed.restapi.dao.ExperimentDao;
import be.iminds.ilabt.jfed.restapi.dao.SpeaksForCredentialDao;
import be.iminds.ilabt.jfed.restapi.debugdata.ClientApiCallDDCollector;
import be.iminds.ilabt.jfed.restapi.debugdata.DebugDataManager;
import be.iminds.ilabt.jfed.restapi.debugdata.ExperimentDDBuilder;
import be.iminds.ilabt.jfed.restapi.debugdata.ServerResourceMethodDDBuilder;
import be.iminds.ilabt.jfed.restapi.dropwizard.JFedRestApiSiftingAppenderFactory;
import be.iminds.ilabt.jfed.restapi.jfedclient.JFedExperimenterApi;
import be.iminds.ilabt.jfed.restapi.jfedclient.JFedExperimenterApiFactory;
import be.iminds.ilabt.jfed.restapi.jfedclient.JFedExperimenterApiImpl;
import be.iminds.ilabt.jfed.restapi.json.Experiment;
import be.iminds.ilabt.jfed.restapi.json.ExperimentBuilder;
import be.iminds.ilabt.jfed.restapi.json.SpeaksForCredential;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.codahale.metrics.annotation.Timed;
import io.swagger.annotations.*;
import jersey.repackaged.com.google.common.cache.CacheBuilder;
import jersey.repackaged.com.google.common.cache.CacheLoader;
import jersey.repackaged.com.google.common.cache.LoadingCache;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;

import javax.inject.Singleton;
import javax.servlet.http.HttpServletRequest;
import javax.validation.constraints.NotNull;
import javax.ws.rs.*;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.sql.Timestamp;
import java.time.Instant;
import java.time.ZonedDateTime;
import java.util.Collections;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

/**
 * MainResource
 */
@SwaggerDefinition(
        info = @Info(
                description = "This API allows you to allocate, provision and delete resources that are available through the GENI AM API.",
                version = "v2.0",
                title = "jFed REST API",
                license = @License(
                        name = "MIT",
                        url = "http://jfed.iminds.be/about/#license"
                )
        ),
        schemes = {SwaggerDefinition.Scheme.HTTPS},
        tags = {
                @Tag(name = "experiment", description = "Operations to create, update and delete experiments"),
                @Tag(name = "speaksfor", description = "Register speaksfor credentials"),
                @Tag(name = "user", description = "Access user information"),
        },
        externalDocs = @ExternalDocs(value = "About the jFed REST API", url = "http://jfed.iminds.be/restapi/")
)


@Path("/experiment")
@Produces(MediaType.APPLICATION_JSON)
@Api(value = "experiment")
@Singleton //because of cache
public class ExperimentResource {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentResource.class);
//    private final JFedTrustStore trustStore;

    @Context
    UriInfo uriInfo;

    private final JFedRestApiConfiguration configuration;
    private final ExperimentDao experimentDao;
    private final SpeaksForCredentialDao speaksForCredentialDao;
    private final DebugDataManager debugDataManager;

    private final SpeakForCredentialParser speakForCredentialParser;
    private final JFedExperimenterApiFactory jFedExperimenterApiFactory;

    private final Map<String, JFedRestApiSiftingAppenderFactory> jFedRestApiAppenderFactoryByMDCKey;

    LoadingCache<Long, Experiment> experimentCache = CacheBuilder.newBuilder()
            .maximumSize(100)
            .expireAfterWrite(10, TimeUnit.MINUTES)
            .build(
                    new CacheLoader<Long, Experiment>() {
                        public Experiment load(Long experimentId) {
                            assert experimentId != null;
                            LOG.debug("CACHE MISS: Fetching Experiment " + experimentId + " from DB");
                            Experiment experiment = experimentDao.findById(experimentId);
                            if (experiment == null)
                                throw new NotFoundException("No Experiment found with id " + experimentId);
                            if (!experiment.getId().equals(experimentId))
                                throw new WebApplicationException("Experiment retrieved for id " + experimentId +
                                        " has mismatched id " + experiment.getId(), Response.Status.INTERNAL_SERVER_ERROR);
                            if (experiment.getSliceUrn() == null)
                                throw new WebApplicationException("No Slice in stored Experiment with id " + experimentId,
                                        Response.Status.INTERNAL_SERVER_ERROR);
                            if (experiment.getSpeaksForCredential() == null)
                                throw new WebApplicationException("No speaks_for credential in stored " +
                                        "Experiment with id " + experimentId, Response.Status.INTERNAL_SERVER_ERROR);

                            return experiment;
                        }
                    });

    private ClientApiCallDDCollector jFedExperimenterApiCacheClientApiCallDDCollection;
    //using this is a bit of a hack, but it works well enough for now
    private LoadingCache<Long, JFedExperimenterApi> jFedExperimenterApiCache = CacheBuilder.newBuilder()
            .maximumSize(100)
            .expireAfterWrite(60, TimeUnit.MINUTES)
            .build(
                    new CacheLoader<Long, JFedExperimenterApi>() {
                        public JFedExperimenterApi load(Long experimentId) throws Exception {
                            LOG.debug("CACHE MISS: CREATING NEW JFedExperimenterApi for Experiment with id " + experimentId);

                            Experiment experiment;
                            try {
                                experiment = experimentCache.get(experimentId);
                            } catch (ExecutionException e) {
                                if (e.getCause() != null) {
                                    if (e.getCause() instanceof Exception) {
                                        throw (Exception) e.getCause();
                                    } else {
                                        throw e;
                                    }
                                } else
                                    throw e;
                            }

                            if (experiment.getSpeakingFor() == null) {
                                throw new WebApplicationException("No speaks_for credential in stored Experiment with id " + experimentId, Response.Status.INTERNAL_SERVER_ERROR);
                            }
                            if (experiment.getSpeaksForCredential() == null) {
                                throw new WebApplicationException("No speaks_for_user in stored Experiment with id " + experimentId, Response.Status.INTERNAL_SERVER_ERROR);
                            }

                            AnyCredential speaksForCredential = speakForCredentialParser.parse("s4cred", experiment.getSpeaksForCredential());
                            assert speaksForCredential != null;

                            JFedExperimenterApi jFedExperimenterApi;
                            try {
                                jFedExperimenterApi = jFedExperimenterApiFactory.create(speaksForCredential, experiment.getSpeakingFor());
                            } catch (IOException e) {
                                LOG.error("Server config error: login PEM file not found: \"" + configuration.getSfaUserLoginPemFilename() + "\"", e);
                                throw new WebApplicationException("Server config error: login PEM file not found.", Response.Status.INTERNAL_SERVER_ERROR);
                            }

                            if (jFedExperimenterApiCacheClientApiCallDDCollection != null)
                                jFedExperimenterApi.setClientApiCallDDCollector(jFedExperimenterApiCacheClientApiCallDDCollection);

                            jFedExperimenterApi.recoverSlice(experiment.getSliceUrn());
                            return jFedExperimenterApi;
                        }
                    });

    public ExperimentResource(JFedRestApiConfiguration configuration,
                              ExperimentDao experimentDao,
                              SpeaksForCredentialDao speaksForCredentialDao,
                              DebugDataManager debugDataManager,
                              Map<String, JFedRestApiSiftingAppenderFactory> jFedRestApiAppenderFactoryByMDCKey,
                              SpeakForCredentialParser speakForCredentialParser,
                              JFedExperimenterApiFactory jFedExperimenterApiFactory) {
        this.configuration = configuration;
        this.experimentDao = experimentDao;
        this.speaksForCredentialDao = speaksForCredentialDao;
        this.debugDataManager = debugDataManager;
        this.jFedRestApiAppenderFactoryByMDCKey = jFedRestApiAppenderFactoryByMDCKey;
        this.speakForCredentialParser = speakForCredentialParser;
        this.jFedExperimenterApiFactory = jFedExperimenterApiFactory;

        if (configuration.getSfaUserLoginPemFilename() == null) {
            LOG.error("Server config error: sfaUserLoginPemFilename option missing");
            throw new WebApplicationException("Server config error: sfaUserLoginPemFilename option missing", Response.Status.INTERNAL_SERVER_ERROR);
        }

        //only needed if trustStore is used for authentication checks (note that these are not required for security as AM does these checks anyway)
//        try {
//            //initialise a JFedExperimenterApi just so we can get the jFed trustStore
//            JFedExperimenterApi jFedExperimenterApi = new JFedExperimenterApi(new File(configuration.getSfaUserLoginPemFilename()), null, null, null);
//            trustStore = jFedExperimenterApi.getJFedTrustStore();
//        } catch (IOException e) {
//            throw new RuntimeException("Failed to initialise JFedExperimenterApi");
//        }
    }

    @POST
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    @ApiOperation(value = "Create an experiment")
    @ApiResponses(value = {
            @ApiResponse(code = 403, message = "You do not properly authenticate"),
            @ApiResponse(code = 400, message = "No speaksfor credential is available to perform your request")
    })
    @Timed
    public Experiment createExperiment(@Context HttpServletRequest request, @NotNull Experiment experiment,
                                       @QueryParam("full") @ApiParam(hidden = true) Boolean full) {
        configuration.assureAccessAllowed(JFedRestApiConfiguration.Access.REST_API, request);

        GeniUrn authenticatedUser = configuration.getAuthenticatedUser(request);
        if (authenticatedUser == null) {
            LOG.debug("FORBIDDING ACCESS: not authenticated");
            throw new WebApplicationException("The SSL connection did not correctly authenticate you as a trusted user. Did you use SSL client authentication?",
                    Response.Status.FORBIDDEN);
        }

        ServerResourceMethodDDBuilder serverResourceMethodDD = new ServerResourceMethodDDBuilder();
        serverResourceMethodDD.setMethod("POST");
        serverResourceMethodDD.setStart(new Timestamp(System.currentTimeMillis()));
        debugDataManager.registerCallStart(serverResourceMethodDD);

        //check minimal input
        if (experiment == null)
            throw new WebApplicationException("Must provide Experiment object", Response.Status.BAD_REQUEST);
        if (experiment.getRequestRspec() == null)
            throw new WebApplicationException("Must provide Experiment object with requestRspec", Response.Status.BAD_REQUEST);

        GeniUrn experimentOwner = authenticatedUser;
        if (experiment.getOwner() != null) {
            experimentOwner = experiment.getOwner();

            authenticatedUser = configuration.modifyAuthenticatedUser(request, authenticatedUser, experimentOwner);
        }
        if (!experimentOwner.equals(authenticatedUser) && !configuration.hasAccess(JFedRestApiConfiguration.Access.ADMIN, authenticatedUser)) {
            LOG.debug("FORBIDDING ACCESS: You are authenticated as user " + authenticatedUser + ". You cannot create an Experiment owned by user " + experimentOwner);
            throw new WebApplicationException("You are authenticated as user " + authenticatedUser + ". You cannot create an Experiment owned by user " + experimentOwner,
                    Response.Status.FORBIDDEN);
        }


        String speaksForCredXmlString = experiment.getSpeaksForCredential();
        GeniUrn serviceUser = configuration.getUser().getUserUrn();
        if (speaksForCredXmlString != null) {
            AnyCredential anyCredential;
            try {
                anyCredential =
                        speakForCredentialParser.parse("credential received in create experiment call from " + authenticatedUser, speaksForCredXmlString);
            } catch (CredentialException e) {
                LOG.error("Error parsing credential", e);
                throw new WebApplicationException("Error parsing credential", Response.Status.BAD_REQUEST);
            }

            GeniUrn spokenForUser =
                    SpeaksForCredentialResource.checkSpeaksForCredential(anyCredential, request, configuration);

            if (!experimentOwner.equals(spokenForUser) && !configuration.hasAccess(JFedRestApiConfiguration.Access.ADMIN, request)) {
                LOG.debug("FORBIDDING ACCESS: You are authenticated as user " + authenticatedUser + " " +
                        "and are creating an experiment as "+experimentOwner+". " +
                        "You cannot use a speaksFor credential for " + spokenForUser);
                throw new WebApplicationException("You are authenticated as user " + authenticatedUser + " " +
                        "and are creating an experiment as "+experimentOwner+". " +
                        "You cannot use a speaksFor credential for " + spokenForUser,
                        Response.Status.FORBIDDEN);
            }

            speaksForCredentialDao.upsert(serviceUser, spokenForUser, anyCredential.getExpiresDate(), speaksForCredXmlString);

            //note: upsert might not replace the credential if one that epxires later already exists.
            //      In that case, this does actually use the credential providied in this call instead of the one in the DB!
        } else {
            SpeaksForCredential res = speaksForCredentialDao.find(serviceUser, experimentOwner);

            if (res == null) {
                LOG.debug("BAD_REQUEST: no speaksfor know or in this request.");
                throw new WebApplicationException(
                        "Must provide Experiment object with speaksForCredential if no speaksfor credential has been registered",
                        Response.Status.BAD_REQUEST);
            } else {
                LOG.debug("No speaksfor credential was provided, but there was one in the DB. It will be used.");
            }

            speaksForCredXmlString = res.getCredentialXml();
            assert speaksForCredXmlString != null;

            //put credential in Experiment object
            experiment = new ExperimentBuilder(experiment).setSpeaksForCredential(speaksForCredXmlString).create();
        }

//        if (experiment.getName() == null)
//            throw new  WebApplicationException("Must provide Experiment object with name", Response.Status.BAD_REQUEST);

        Instant expire;
        if (experiment.getExpire() != null) {
            expire = experiment.getExpire().toInstant();
        } else {
            //fall back to default
            expire = ZonedDateTime.now()
                    .plusDays(0)
                    .plusHours(2)
                    .plusMinutes(0)
                    .plusSeconds(0).toInstant(); //=expiration time
        }

        AnyCredential s4cred;
        try {
            assert speaksForCredXmlString != null;
            s4cred = speakForCredentialParser.parse("s4cred", speaksForCredXmlString);
        } catch (CredentialException e) {
            LOG.error("Experiment.speaksForCredential is not a valid credential", e);
            throw new WebApplicationException("Experiment.speaksForCredential is not a valid credential: " + e.getMessage(), Response.Status.BAD_REQUEST);
        }
        GeniUrn spokenForUser = SpeaksForCredentialResource.checkSpeaksForCredential(s4cred, request, configuration);
        assert authenticatedUser != null;
        assert experimentOwner != null;
        assert spokenForUser.equals(experimentOwner);

        LOG.debug("Caller is authorized");

        ExperimentBuilder expBuilder = new ExperimentBuilder(experiment);
        expBuilder.setOwner(experimentOwner);
        expBuilder.setCreated(new Timestamp(System.currentTimeMillis()));
        Long newExperimentId = experimentDao.insert(expBuilder.create());
        expBuilder.setId(newExperimentId);
        MDC.put("experimentId", newExperimentId + "");
        ExperimentDDBuilder experimentDD = new ExperimentDDBuilder(expBuilder.create());
        experimentDD.addServerResourceMethodDD(serverResourceMethodDD);
        debugDataManager.addExperiment(experimentDD);
        try {

            experimentCache.put(newExperimentId, expBuilder.create());

            JFedExperimenterApi jFedExperimenterApi;
            try {
                jFedExperimenterApi =
                        new JFedExperimenterApiImpl(new File(configuration.getSfaUserLoginPemFilename()), null, Collections.singletonList(s4cred), spokenForUser);
            } catch (IOException e) {
                LOG.error("Server config error: login PEM file not found: \"" + configuration.getSfaUserLoginPemFilename() + "\"", e);
                throw new WebApplicationException("Server config error: login PEM file not found.", Response.Status.INTERNAL_SERVER_ERROR);
            } catch (SimpleGeniUserWithSpeaksFor.InvalidSpeaksForException e) {
                LOG.error("Speaksfor credential invalid", e);
                throw new WebApplicationException("Invalid speaksfor credential: " + e.getMessage(), e, Response.Status.BAD_REQUEST);
            }

            jFedExperimenterApi.setClientApiCallDDCollector(serverResourceMethodDD);

            //use jFed client
//        String newSliceName = experiment.getName();
//        assert newSliceName != null;
            String newSliceName;
            if (experiment.getSliceName() != null) {
                newSliceName = experiment.getSliceName();
            } else {
                newSliceName = "";//experiment.getName();
                if (newSliceName.length() > 4)
                    newSliceName = newSliceName.replaceAll("[aeuoiAEUOI]", "").substring(0, 4);
                newSliceName += randChars(7 - newSliceName.length());
            }

            expBuilder.setStatus(Experiment.Status.STARTING);
            expBuilder.setSpeakingFor(spokenForUser);
            expBuilder.setSliceName(newSliceName);

            experimentCache.put(newExperimentId, expBuilder.create());

            assert experiment.getProject() != null;
            //TODO automatically choose a project
            jFedExperimenterApi.createSliceAndSlivers(newSliceName, experiment.getProject(), experiment.getRequestRspec(), expire);
            assert jFedExperimenterApi.getExperimentController() != null;
            assert jFedExperimenterApi.getExperiment() != null;
            assert jFedExperimenterApi.getExperimentSlice() != null;
            assert jFedExperimenterApi.getSliceUrn() != null;

            expBuilder.setProject(jFedExperimenterApi.getProject());
            expBuilder.setSliceName(jFedExperimenterApi.getSliceName());
            expBuilder.setSliceUrn(jFedExperimenterApi.getSliceUrn());
            expBuilder.setExpire(jFedExperimenterApi.getExpire());
            expBuilder.setManifestRspec(jFedExperimenterApi.getManifestRspec());

            Experiment dbRes = expBuilder.create();
            experimentDao.update(dbRes);

            experimentCache.put(newExperimentId, dbRes);
            jFedExperimenterApiCache.put(newExperimentId, jFedExperimenterApi);

            if (full == null || !full) {
                expBuilder.setRequestRspec(null);
                expBuilder.setSpeaksForCredential(null);
            }
            addUri(expBuilder);
            Experiment res = expBuilder.create();
            return res;
        } catch (Throwable t) {
            LOG.error("Exception handling " + getClass().getSimpleName() + " call", t);
            throw new RuntimeException("Exception creating experiment", t);
//            throw new ServerErrorException("Exception creating experiment", Response.Status.INTERNAL_SERVER_ERROR, t);
        } finally {
            //do not remove: all calls should override MDC,
            // if anything later uses this, it might belong to this experiment, so we log it as such
//            MDC.remove("experimentId");
            serverResourceMethodDD.setStop(new Timestamp(System.currentTimeMillis()));
            debugDataManager.registerCallEnd(serverResourceMethodDD, experimentDD);
            debugDataManager.save(experimentDD);
        }
    }

    private static String randCharSet = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
    private static Random rand = new Random();

    private static final String randChars(int len) {
        String res = "";
        for (int i = 0; i < len; i++) {
            int r = rand.nextInt(randCharSet.length());
            res += randCharSet.charAt(r);
        }
        return res;
    }

    private RuntimeException getRethrowableExecutionExceptionCause(ExecutionException e, String failedActionDescription) {
        if (e.getCause() != null) {
            if (e.getCause() instanceof RuntimeException) {
                return (RuntimeException) e.getCause();
            } else {
                LOG.error(failedActionDescription, e.getCause());
                return new WebApplicationException(failedActionDescription + ": " + e.getCause().getMessage(), e.getCause(), Response.Status.INTERNAL_SERVER_ERROR);
            }
        } else
            return new WebApplicationException(failedActionDescription + ": " + e.getMessage(), e, Response.Status.INTERNAL_SERVER_ERROR);
    }

    @GET
    @Path("{id}")
    @Produces(MediaType.APPLICATION_JSON)
    @ApiOperation(value = "Get the current status of an experiment", notes = "This action requests and returns the current status of all resources in the experiment")
    @Timed
    public Experiment getExperimentStatus(@NotNull @PathParam("id") Long id, @Context HttpServletRequest request,
                                          @QueryParam("full") @ApiParam(hidden = true) Boolean full) {
        configuration.assureAccessAllowed(JFedRestApiConfiguration.Access.REST_API, request);
        assert id != null;

        ServerResourceMethodDDBuilder serverResourceMethodDD = new ServerResourceMethodDDBuilder();
        serverResourceMethodDD.setMethod("GET");
        serverResourceMethodDD.setStart(new Timestamp(System.currentTimeMillis()));
        debugDataManager.registerCallStart(serverResourceMethodDD);

        Experiment experiment;
        try {
            experiment = experimentCache.get(id);
        } catch (ExecutionException e) {
            throw getRethrowableExecutionExceptionCause(e, "Error getting Experiment with ID " + id);
        }
        assert experiment != null;
        assert experiment.getSliceUrn() != null;
        assert experiment.getId().equals(id);

        ExperimentDDBuilder experimentDD = debugDataManager.getExperiment(experiment);
        experimentDD.setFromExperiment(experiment);
        experimentDD.addServerResourceMethodDD(serverResourceMethodDD);

        MDC.put("experimentId", experiment.getId() + "");
        try {
            checkAuthorization(experiment, request);

            JFedExperimenterApi jFedExperimenterApi;
            try {
                jFedExperimenterApiCacheClientApiCallDDCollection = serverResourceMethodDD;
                jFedExperimenterApi = jFedExperimenterApiCache.get(experiment.getId());
            } catch (ExecutionException e) {
                throw new WebApplicationException("Failed to recover slice " + experiment.getSliceUrn(), e, Response.Status.INTERNAL_SERVER_ERROR);
            }

            jFedExperimenterApi.setClientApiCallDDCollector(serverResourceMethodDD);

            //refreshStatusIfNeeded() will not refresh status if not ready yet, or too soon after previous refresh.
            //refreshStatusIfNeeded() might block for up to 5 seconds, but not longer.
            jFedExperimenterApi.refreshStatusIfNeeded();

            Experiment res = syncDBExperimentToClient(jFedExperimenterApi, experiment, full == null ? false : full);
            return addUri(res);
        } catch (Throwable t) {
            LOG.error("Exception handling " + getClass().getSimpleName() + " call", t);
            throw t;
        } finally {
            //do not remove: all calls should override MDC,
            // if anything later uses this, it might belong to this experiment, so we log it as such
//            MDC.remove("experimentId");
            serverResourceMethodDD.setStop(new Timestamp(System.currentTimeMillis()));
            debugDataManager.registerCallEnd(serverResourceMethodDD, experimentDD);
            debugDataManager.save(experimentDD);
        }
    }

    @DELETE
    @Path("{id}")
    @Produces(MediaType.APPLICATION_JSON)
    @ApiOperation(value = "Delete an experiment", notes = "This action will destroy all the resources currently provisioned in the experiment.")
    @Timed
    public Experiment deleteExperiment(@NotNull @PathParam("id") Long id, @Context HttpServletRequest request,
                                       @QueryParam("full") @ApiParam(hidden = true) Boolean full) {
        configuration.assureAccessAllowed(JFedRestApiConfiguration.Access.REST_API, request);
        assert id != null;

        ServerResourceMethodDDBuilder serverResourceMethodDD = new ServerResourceMethodDDBuilder();
        serverResourceMethodDD.setMethod("DELETE");
        serverResourceMethodDD.setStart(new Timestamp(System.currentTimeMillis()));
        debugDataManager.registerCallStart(serverResourceMethodDD);

        Experiment experiment;
        try {
            experiment = experimentCache.get(id);
        } catch (ExecutionException e) {
            throw getRethrowableExecutionExceptionCause(e, "Error getting Experiment with ID " + id);
        }
        assert experiment != null;
        assert experiment.getSliceUrn() != null;
        assert experiment.getId() == id;

        ExperimentDDBuilder experimentDD = debugDataManager.getExperiment(experiment);
        experimentDD.setFromExperiment(experiment);
        experimentDD.addServerResourceMethodDD(serverResourceMethodDD);

        MDC.put("experimentId", experiment.getId() + "");
        try {
            checkAuthorization(experiment, request);

            JFedExperimenterApi jFedExperimenterApi;
            try {
                jFedExperimenterApiCacheClientApiCallDDCollection = serverResourceMethodDD;
                jFedExperimenterApi = jFedExperimenterApiCache.get(experiment.getId());
            } catch (ExecutionException e) {
                throw new WebApplicationException("Failed to recover slice " + experiment.getSliceUrn(), Response.Status.INTERNAL_SERVER_ERROR);
            }

            jFedExperimenterApi.setClientApiCallDDCollector(serverResourceMethodDD);

            jFedExperimenterApi.deleteSlice();

            Experiment res = syncDBExperimentToClient(jFedExperimenterApi, experiment, full == null ? false : full);
            return addUri(res);
        } catch (Throwable t) {
            LOG.error("Exception handling " + getClass().getSimpleName() + " call", t);
            throw t;
        } finally {
            //do not remove: all calls should override MDC,
            // if anything later uses this, it might belong to this experiment, so we log it as such
//            MDC.remove("experimentId");
            serverResourceMethodDD.setStop(new Timestamp(System.currentTimeMillis()));
            debugDataManager.registerCallEnd(serverResourceMethodDD, experimentDD);
            debugDataManager.save(experimentDD);
        }
    }

    @PUT
    @Path("{id}")
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    @ApiOperation(value = "Change the expiration date of an experiment", notes = "When an experiment expires, all resources are released. To prevent this from happening, you can move the expiration date to a time farther in the future.")
    @ApiResponses(
            @ApiResponse(code = 400, message = "An invalid experiment id was submitted, or one or more immutable fields of the experiment-object were changed")
    )
    @Timed
    public Experiment renewExperiment(@NotNull @PathParam("id") Long id, @Context HttpServletRequest request,
                                      @QueryParam("full") @ApiParam(hidden = true) Boolean full,
                                      Experiment updatedExperiment) {
        configuration.assureAccessAllowed(JFedRestApiConfiguration.Access.REST_API, request);
        assert id != null;

        ServerResourceMethodDDBuilder serverResourceMethodDD = new ServerResourceMethodDDBuilder();
        serverResourceMethodDD.setMethod("PUT");
        serverResourceMethodDD.setStart(new Timestamp(System.currentTimeMillis()));
        debugDataManager.registerCallStart(serverResourceMethodDD);

        if (updatedExperiment.getId() == null)
            throw new WebApplicationException("Provided experiment must have ID. (path ID=" + id + ")",
                    Response.Status.BAD_REQUEST);
        if (updatedExperiment.getId() != id)
            throw new WebApplicationException("Provided experiment does not have correct ID. path ID=" + id + " experiment ID=" + updatedExperiment.getId(),
                    Response.Status.BAD_REQUEST);

        Experiment experiment;
        try {
            experiment = experimentCache.get(id);
        } catch (ExecutionException e) {
            throw getRethrowableExecutionExceptionCause(e, "Error getting Experiment with ID " + id);
        }
        assert experiment != null;
        assert experiment.getSliceUrn() != null;
        assert experiment.getId() == id;

        ExperimentDDBuilder experimentDD = debugDataManager.getExperiment(experiment);
        experimentDD.setFromExperiment(experiment);
        experimentDD.addServerResourceMethodDD(serverResourceMethodDD);

        MDC.put("experimentId", experiment.getId() + "");
        try {
            checkAuthorization(experiment, request);

//        if (updatedExperiment.getName() != null && !updatedExperiment.getName().equals(experiment.getName()))
//            throw new  WebApplicationException("Cannot change name of experiment (can only change expire date).",
//                    Response.Status.BAD_REQUEST);
            if (updatedExperiment.getAm() != null && !updatedExperiment.getAm().equals(experiment.getAm()))
                throw new WebApplicationException("Cannot change am of experiment (can only change expire date).",
                        Response.Status.BAD_REQUEST);
            if (updatedExperiment.getCreated() != null && !updatedExperiment.getCreated().equals(experiment.getCreated()))
                throw new WebApplicationException("Cannot change create date of experiment (can only change expire date).",
                        Response.Status.BAD_REQUEST);
            if (updatedExperiment.getOwner() != null && !updatedExperiment.getOwner().equals(experiment.getOwner()))
                throw new WebApplicationException("Cannot change owner of experiment (can only change expire date).",
                        Response.Status.BAD_REQUEST);
            if (updatedExperiment.getSharedWith() != null && !updatedExperiment.getSharedWith().equals(experiment.getSharedWith()))
                throw new WebApplicationException("Cannot change sharedWith of experiment (can only change expire date).",
                        Response.Status.BAD_REQUEST);
            if (updatedExperiment.getSliceName() != null && !updatedExperiment.getSliceName().equals(experiment.getSliceName()))
                throw new WebApplicationException("Cannot change slice name of experiment (can only change expire date).",
                        Response.Status.BAD_REQUEST);
            if (updatedExperiment.getSliceUrn() != null && !updatedExperiment.getSliceUrn().equals(experiment.getSliceUrn()))
                throw new WebApplicationException("Cannot change slice of experiment (can only change expire date).",
                        Response.Status.BAD_REQUEST);
            if (updatedExperiment.getProject() != null && !updatedExperiment.getProject().equals(experiment.getProject()))
                throw new WebApplicationException("Cannot change project of experiment (can only change expire date).",
                        Response.Status.BAD_REQUEST);

            Timestamp newExpire = updatedExperiment.getExpire();
            assert experiment.getExpire() != null;
            if (newExpire == null)
                throw new WebApplicationException("Can only change expire date of experiment, but none was provided.",
                        Response.Status.BAD_REQUEST);
            if (experiment.getExpire() != null) {
                if (newExpire.equals(experiment.getExpire()))
                    throw new WebApplicationException("New expire date of experiment was the same as current (so nothing to do).",
                            Response.Status.BAD_REQUEST);
                if (newExpire.before(experiment.getExpire()))
                    throw new WebApplicationException("New expire date of experiment may not be before current expire date: can only increase expire date.",
                            Response.Status.BAD_REQUEST);
            } else {
                LOG.error("BUG: experiment.getExpire() == null. Will ignore and try to update expire date.");
            }

            LOG.debug("all checks OK: allowed to Renew slice and slivers");

            JFedExperimenterApi jFedExperimenterApi;
            try {
                jFedExperimenterApiCacheClientApiCallDDCollection = serverResourceMethodDD;
                jFedExperimenterApi = jFedExperimenterApiCache.get(experiment.getId());
            } catch (ExecutionException e) {
                throw new WebApplicationException("Failed to recover slice " + experiment.getSliceUrn(), Response.Status.INTERNAL_SERVER_ERROR);
            }

            jFedExperimenterApi.setClientApiCallDDCollector(serverResourceMethodDD);

            jFedExperimenterApi.renewSlice(newExpire);

            Experiment res = syncDBExperimentToClient(jFedExperimenterApi, experiment, full == null ? false : full);
            return addUri(res);
        } catch (Throwable t) {
            LOG.error("Exception handling " + getClass().getSimpleName() + " call", t);
            throw t;
        } finally {
            //do not remove: all calls should override MDC,
            // if anything later uses this, it might belong to this experiment, so we log it as such
//            MDC.remove("experimentId");
            serverResourceMethodDD.setStop(new Timestamp(System.currentTimeMillis()));
            debugDataManager.registerCallEnd(serverResourceMethodDD, experimentDD);
        }
    }

    public ExperimentBuilder addUri(ExperimentBuilder experimentBuilder) {
        String base =
                configuration == null || configuration.getBaseUrl() == null || configuration.getBaseUrl().trim().isEmpty()
                        ?
                        null : configuration.getBaseUrl();
        if (base != null && !base.endsWith("/")) base += "/";

        URI uri = null;
        try {
            uri = new URI(base + "experiment/" + experimentBuilder.getId());
            experimentBuilder.setUri(uri);
        } catch (URISyntaxException e) {
            LOG.error("Error when adding URI", e);
            //throw new WebApplicationException("Error when adding URI", Response.Status.INTERNAL_SERVER_ERROR);
            //ignore
        }

        return experimentBuilder;
    }

    public Experiment addUri(Experiment experiment) {
        return addUri(new ExperimentBuilder(experiment)).create();
    }

    private static Experiment.Status translateState(be.iminds.ilabt.jfed.experiment.ExperimentState state) {
        switch (state) {
            case PENDING:
                return Experiment.Status.STARTING;
            case RESTORING:
                return Experiment.Status.STARTING;
            case CREATED:
                return Experiment.Status.STARTING;
            case RESERVING:
                return Experiment.Status.STARTING;
            case FUTURE_RESERVATION:
                return Experiment.Status.NONEXISTING;
            case PROVISIONING:
                return Experiment.Status.STARTING;
            case TESTING:
                return Experiment.Status.READY; //normally, this state will be automatically skipped because no experimentTester is set
            case READY:
                return Experiment.Status.READY;
            case DELETING:
                return Experiment.Status.NONEXISTING;
            case EXPIRING:
                return Experiment.Status.STARTING;
            case EXPIRED:
                return Experiment.Status.NONEXISTING;
            case FAILED:
                return Experiment.Status.FAILED;
            case EMPTY:
                return Experiment.Status.NONEXISTING;
            default:
                throw new RuntimeException("Unsupported ExperimentState: " + state);
        }
    }

    private static Experiment.Status translateState(StatusDetails state) {
        switch (state.getGlobalStatus()) {
            case UNINITIALISED:
                return Experiment.Status.NONEXISTING;
            case READY:
                return Experiment.Status.READY;
            case UNALLOCATED:
                return Experiment.Status.NONEXISTING;
            case UNKNOWN:
                return Experiment.Status.STARTING;
            case CHANGING:
                return Experiment.Status.STARTING;
            case FAIL:
                return Experiment.Status.FAILED;
            default:
                throw new RuntimeException("Unsupported ExperimentState: " + state);
        }
    }

    private Experiment syncDBExperimentToClient(JFedExperimenterApi jFedExperimenterApi, Experiment dbExperiment, boolean full) {
        ExperimentBuilder updatedDbExperimentBuilder = new ExperimentBuilder(dbExperiment);
        updatedDbExperimentBuilder.setProject(jFedExperimenterApi.getProject());
        updatedDbExperimentBuilder.setSliceName(jFedExperimenterApi.getSliceName());
        updatedDbExperimentBuilder.setSliceUrn(jFedExperimenterApi.getSliceUrn());
        updatedDbExperimentBuilder.setExpire(jFedExperimenterApi.getExpire());
        updatedDbExperimentBuilder.setManifestRspec(jFedExperimenterApi.getManifestRspec());

        //start with State from Slice object
        Experiment.Status sliceStatus;
        if (jFedExperimenterApi.getExperimentSliceStatus() != null) {
            sliceStatus = translateState(jFedExperimenterApi.getExperimentSliceStatus());
        } else {
            sliceStatus = Experiment.Status.NONEXISTING;
        }
        Experiment.Status newStatus = sliceStatus;

        //update with state from Experiment in some cases
        if (jFedExperimenterApi.getExperimentState() == null) {
            //don't change newStatus
        } else {
            switch (jFedExperimenterApi.getExperimentState()) {
                case PENDING: {
                    newStatus = Experiment.Status.STARTING;
                    break;
                }
                case RESTORING: {
                    newStatus = Experiment.Status.STARTING;
                    break;
                }
                case CREATED: {
                    newStatus = Experiment.Status.STARTING;
                    break;
                }
                case RESERVING:
                    break;
                case FUTURE_RESERVATION:
                    break;
                case PROVISIONING: {
                    newStatus = Experiment.Status.STARTING;
                    break;
                }
                case TESTING:
                    break;
                case READY:
                    break;
                case DELETING:
                    break;
                case EXPIRING:
                    break;
                case EXPIRED:
                    break;
                case EMPTY:
                    break;
                case FAILED: {
                    newStatus = Experiment.Status.FAILED;
                    break;
                }
                default:
                    throw new RuntimeException("Unsupported ExperimentState: " + jFedExperimenterApi.getExperimentState());
            }
        }
        updatedDbExperimentBuilder.setStatus(newStatus);
        LOG.debug("Updating Experiment.status: sliceStatus=" + sliceStatus + " " +
                "experimentStatus=" + jFedExperimenterApi.getExperimentState() + " " +
                "combined=" + newStatus);

        if (jFedExperimenterApi.getExperimentSliceStatus() != null
                && jFedExperimenterApi.getExperimentSliceStatus().getGlobalStatus() != null) {
            updatedDbExperimentBuilder.setGlobalSliverStatus(jFedExperimenterApi.getExperimentSliceStatus().getGlobalStatus().name());
        }
        if (jFedExperimenterApi.getExperimentState() != null) {
            updatedDbExperimentBuilder.setJFedExperimentState(jFedExperimenterApi.getExperimentState().name());
        }


        Experiment dbExp = updatedDbExperimentBuilder.create();
        experimentDao.update(dbExp);
        experimentCache.put(dbExp.getId(), dbExp);

        if (!full) {
            updatedDbExperimentBuilder.setRequestRspec(null);
            updatedDbExperimentBuilder.setSpeaksForCredential(null);
        }
        Experiment res = updatedDbExperimentBuilder.create();

        return res;
    }


    private void checkAuthorization(Experiment experiment, HttpServletRequest request) {
        //check if requesting user is owner (or in sharedWith) of the experiment

        assert experiment != null;
        assert experiment.getOwner() != null;

        GeniUrn authenticatedUser = configuration.getAuthenticatedUser(request);
        if (authenticatedUser == null) {
            throw new WebApplicationException("The SSL connection did not correctly authenticate you as a trusted user. " +
                    "Did you use SSL client authentication?",
                    Response.Status.FORBIDDEN);
        }

        if (configuration.hasAccess(JFedRestApiConfiguration.Access.SPEAK_FOR_OTHERS, request))
            return;
        if (configuration.hasAccess(JFedRestApiConfiguration.Access.ADMIN, request))
            return;

        if (authenticatedUser.equals(experiment.getOwner())) {
            return;
        }

        if (experiment.getSharedWith() != null && !experiment.getSharedWith().isEmpty()) {
            for (GeniUrn sharedWithUser : experiment.getSharedWith()) {
                if (authenticatedUser.equals(sharedWithUser)) {
                    return;
                }
            }
        }

        LOG.error("Experiment access denied for " + authenticatedUser + " to Experiment " + experiment.getId() + " of " +
                experiment.getOwner() + " shared with " + experiment.getSharedWith());
        throw new WebApplicationException("The user making this call has no rights to access this Experiment",
                Response.Status.FORBIDDEN);
    }
}
